-- .BMP/.ICO file format wrapper                                  G.Burke 12/04
-- From memory I _think_ this stuff works


include PE_struct.e
include file.e

constant    
	    
	BITMAPFILEHEADER={
			{"Signature"        ,2},
			{"FileSize"         ,4},
			{"Reserved"         ,4},
			{"DataOffset"       ,4}     },

	BITMAPINFOHEADER={
			{"Size"             ,4},
			{"Width"            ,4},
			{"Height"           ,4},
			{"Planes"           ,2},
			{"BitCount"         ,2},
			{"Compression"      ,4},
			{"SizeImage"        ,4},
			{"XpelsPerMeter"    ,4},
			{"YpelsPerMeter"    ,4},
			{"ClrUsed"          ,4},
			{"ClrImportant"     ,4}     },
			
	valid_bmp_bpp=  {1,4,8,16,24,32},

	ICONDIR={
			{"Reserved"         ,2},
			{"Type"             ,2},
			{"Count"            ,2}     },


	ICONDIRENTRY={
			{"Width"            ,1},
			{"Height"           ,1},
			{"ColorCount"       ,1},
			{"Reserved"         ,1},
			{"Planes"           ,2},
			{"BitCount"         ,2},
			{"SizeInBytes"      ,4},
			{"FileOffset"       ,4}     }




---------------------------------------------------------------------
-------------------------Read--BMP-----------------------------------
---------------------------------------------------------------------


function line_size(integer w,integer bpp)
    --calculate the size in bytes of a single pixel line including padding
    integer n,r
    if bpp=32 then 
	return w*4
    elsif bpp=24 then
	n=w*3
    elsif bpp=16 then
	n=w*2
    else
	n=floor((w*bpp)/8+.999)
    end if
    r=remainder(n,4)
    return n+(4-r)-4*(not r)
end function

function check_bmp_headers(sequence fh,sequence ih)
    --check for unsupported formats
    -- returns 0 if ok 1 if unsupported
    return (
       compare(PE_get_member(fh,BITMAPFILEHEADER,"Signature"),"BM")+
	      (PE_get_member(ih,BITMAPINFOHEADER,"Size"     )!=40 )+
	      (PE_get_member(ih,BITMAPINFOHEADER,"Planes"   ) >1  )+
	 (find(PE_get_member(ih,BITMAPINFOHEADER,"BitCount" ),valid_bmp_bpp)<1)+
	       PE_get_member(ih,BITMAPINFOHEADER,"Compression") )!=0
end function

function load_pixels(   integer fn,
			integer width,
			integer height,
			integer bpp         )
    atom a,c
    sequence l,r
    r=repeat({},height)
    if bpp=1 then
	a=width/8
	a+= not integer(a) 
	if remainder(a,4) then
	    a+=(4-remainder(a,4))
	end if
	for y=height to 1 by -1 do
	    l={}
	    for x=1 to a do
		c=getc(fn)
		l&=and_bits(c,#80) !=0
		l&=and_bits(c,#40) !=0
		l&=and_bits(c,#20) !=0
		l&=and_bits(c,#10) !=0
		l&=and_bits(c,#08) !=0
		l&=and_bits(c,#04) !=0
		l&=and_bits(c,#02) !=0
		l&=and_bits(c,#01) !=0
	    end for
	    r[y]=l[1..width]
	end for
	return r
    elsif bpp=4 then
	a=width/2
	a+= not integer(a) 
	if remainder(a,4) then
	    a+=(4-remainder(a,4))
	end if
	for y=height to 1 by -1 do
	    l={}
	    for x=1 to a do
		c=getc(fn)
		l&=and_bits(c,#F0)/#10
		l&=and_bits(c,#0F)
	    end for
	    r[y]=l[1..width]
	end for
	return r
    elsif bpp=8 then
	a=width
	if remainder(a,4) then
	    a+=(4-remainder(a,4))
	end if
	for y=height to 1 by -1 do
	    l={}
	    for x=1 to a do
		l&=getc(fn)
	    end for
	    r[y]=l[1..width]
	end for
	return r
    elsif bpp=16 then
	a=width*2
	if remainder(a,4) then
	    a+=(4-remainder(a,4))
	end if
	for y=height to 1 by -1 do
	    l=repeat({},a/2)
	    for x=1 to a/2 do
		l[x]={getc(fn),getc(fn)}
	    end for
	    r[y]=l[1..width]
	end for
	return r
    elsif bpp=24 then
	for y=height to 1 by -1 do
	    l={}
	    l=repeat({},width)
	    for x=1 to width do
		l[x]={getc(fn),getc(fn),getc(fn)}
	    end for
	    r[y]=l[1..width]
	end for
	return r
    elsif bpp=32 then
	for y=height to 1 by -1 do
	    l=repeat({},width)
	    for x=1 to width do
		l[x]={getc(fn),getc(fn),getc(fn),getc(fn)}
	    end for
	    r[y]=l[1..width]
	end for
	return r
    end if
end function

function load_palette(integer fn,integer count)
    sequence r,p
    count+=256*(not count)
    r=repeat({},count)
    for x=1 to count do
	p={}
	for y=1 to 3 do
	    p=getc(fn)&p
	end for
	r[x]=p
	if getc(fn) then end if
    end for
    return r
end function

global function load_bmp_ex(sequence path)
    -- returns  { file hdr, info hdr, palette, pixels }
    integer fn,bpp
    sequence r
    r={} 
    fn=open(path,"rb")
    if fn>1 then
	r=append(r,PE_get_struct_from_file(fn,BITMAPFILEHEADER))
	r=append(r,PE_get_struct_from_file(fn,BITMAPINFOHEADER))
	if check_bmp_headers(r[1],r[2]) then return -1 end if
	bpp=PE_get_member(r[2],BITMAPINFOHEADER,"BitCount")
	if bpp<9 then
	    r=append(r,load_palette(fn,power(2,bpp)))
	else
	    r=append(r,{})
	end if
	
	r=append(r,load_pixels( fn,
			PE_get_member(r[2],BITMAPINFOHEADER,"Width" ),
			PE_get_member(r[2],BITMAPINFOHEADER,"Height"),
			    bpp                                  )  )
	close(fn)
	return r
    else
	return -1
    end if
end function

global function load_bmp(sequence path)
    -- returns {palette,pixels}
    object r
    r=load_bmp_ex(path)
    if sequence(r) then
	return r[3..4]
    else
	return r
    end if
end function

---------------------------------------------------------------------
--------------------------Write--BMP---------------------------------
---------------------------------------------------------------------

procedure write_pixels(     integer fn,
			    integer bpp,
			    sequence pix   )
    atom a
    integer width,height
    sequence l,r
    width=length(pix[1])
    height=length(pix)
    r=repeat({},height)
    if bpp=1 then
	a=width/8
	a+= not integer(a) 
	if remainder(a,4) then
	    a+=(4-remainder(a,4))
	end if
	for y=height to 1 by -1 do
	    l=repeat(0,a)
	    for x=1 to width do
		l[floor((x-1)/8)+1]+=pix[y][x]*power(2,7-remainder((x-1),8))
	    end for
	    puts(fn,l)
	end for
    elsif bpp=4 then
	a=width/2
	a+= not integer(a) 
	if remainder(a,4) then
	    a+=(4-remainder(a,4))
	end if
	for y=height to 1 by -1 do
	    l=repeat(0,a)
	    for x=1 to width do
		l[floor((x-1)/2)+1]+=pix[y][x]*power(16,remainder(x,2))
	    end for
	    puts(fn,l)
	end for
    elsif bpp=8 then
	a=width
	if remainder(a,4) then
	    a+=(4-remainder(a,4))
	end if
	for y=height to 1 by -1 do
	    l=pix[y]&repeat(0,4)
	    puts(fn,l[1..a])
	end for
    elsif bpp=16 then
	a=width*2
	if remainder(a,4) then
	    a+=(4-remainder(a,4))
	end if
	for y=height to 1 by -1 do
	    l={}
	    for x=1 to width do
		l&=pix[y][x]
	    end for
	    l&=repeat(0,4)
	    puts(fn,l[1..a])
	end for
    elsif bpp=24 then
	a=width*3
	if remainder(a,4) then
	    a+=(4-remainder(a,4))
	end if
	for y=height to 1 by -1 do
	    l={}
	    for x=1 to width do
		l&=pix[y][x]
	    end for
	    l&=repeat(0,4)
	    puts(fn,l[1..a])
	end for
    elsif bpp=32 then
	for y=height to 1 by -1 do
	    for x=1 to width do
		puts(fn,pix[y][x])
	    end for
	end for
    end if
end procedure


procedure write_palette(integer fn,sequence p)
    for x=1 to length(p) do
	puts(fn,p[x][3])
	puts(fn,p[x][2])
	puts(fn,p[x][1])
	puts(fn,0)
    end for
end procedure

global function save_bmp(sequence bm,integer bpp,sequence path)
    -- bm is {palette,pixels}
    -- if > 8bit {0,pixels}
    
    sequence fh, -- file header
	     ih  -- info header
    integer fn
    
    fh=repeat(0,length(BITMAPFILEHEADER))
    ih=repeat(0,length(BITMAPINFOHEADER))
    
    fh=PE_set_member(fh,BITMAPFILEHEADER,"Signature", bytes_to_int({'B','M',0,0})) 
    fh=PE_set_member(fh,BITMAPFILEHEADER,"DataOffset",PE_size_of(BITMAPFILEHEADER)+
						   PE_size_of(BITMAPINFOHEADER)+
						   length(bm[1])*4      )
    ih=PE_set_member(ih,BITMAPINFOHEADER,"Size"    ,40)
    ih=PE_set_member(ih,BITMAPINFOHEADER,"Width"   ,length(bm[2][1]))
    ih=PE_set_member(ih,BITMAPINFOHEADER,"Height"  ,length(bm[2]))
    ih=PE_set_member(ih,BITMAPINFOHEADER,"Planes"  ,1)
    ih=PE_set_member(ih,BITMAPINFOHEADER,"BitCount",bpp)
    ih=PE_set_member(ih,BITMAPINFOHEADER,"ClrUsed" ,length(bm[1]))
    
    fn=open(path,"wb")
    if fn>1 then
	close(fn)
	fn=open(path,"ub")
	PE_write_struct_to_file(fn,BITMAPFILEHEADER,fh)
	PE_write_struct_to_file(fn,BITMAPINFOHEADER,ih)
	write_palette(fn,bm[1])
	write_pixels(fn,bpp,bm[2])
	close(fn)
	return 0
    else
	return 1
    end if
end function

---------------------------------------------------------------------
--------------------------ICO-Files----------------------------------
---------------------------------------------------------------------


function get_bc(integer x)
    --returns a bit count from a color count
    return 8*(x=0)+1*(x=2)+4*(x=16)
end function

function get_cc(integer x)
    --returns a color count from a bit count
    return and_bits(power(2,x),#FF)
end function

function fix_ico_entry(sequence s)
    integer bpp,cc
    bpp=PE_get_member(s,ICONDIRENTRY,"BitCount")
    cc =PE_get_member(s,ICONDIRENTRY,"ColorCount")
    if bpp then
	return PE_set_member(s,ICONDIRENTRY,"ColorCount",get_cc(bpp))
    else
	return PE_set_member(s,ICONDIRENTRY,"BitCount"  ,get_bc(cc))
    end if
end function

function fix_bih(sequence bih,sequence ie)
    if not PE_get_member(bih,BITMAPINFOHEADER,"BitCount") then
	return PE_set_member(bih,BITMAPINFOHEADER,"BitCount",
		PE_get_member(ie,ICONDIRENTRY,    "BitCount")  )
    end if
    return bih
end function

global function load_ico(sequence path)
    -- Loads a .ICO file
    -- returns  list of
    --          {   entry header,
    --              BITMAPINFOHEADER,
    --              palettes,
    --              XOR bmp,
    --              AND bmp         }  entries
    sequence r
    integer fn
    fn=open(path,"rb")
    if fn>1 then
	-- ICO Header
	r=PE_get_struct_from_file(fn,ICONDIR)
	
	if PE_get_member(r,ICONDIR,"Type")!=1 then return -1 end if
	
	-- ICO Entry Headers
	r=repeat({},PE_get_member(r,ICONDIR,"Count"))
	for x=1 to length(r) do
	    r[x]=append(r[x],fix_ico_entry(PE_get_struct_from_file(fn,ICONDIRENTRY)))
	end for
	for x=1 to length(r) do
	    --bitmap info headers
	    if seek(fn,PE_get_member(r[x][1],ICONDIRENTRY,"FileOffset")) then end if
	    r[x]=append(r[x],fix_bih(PE_get_struct_from_file(fn,BITMAPINFOHEADER),
					    r[x][1]))

	    -- get palette if 8-bit or less
	    
	    if PE_get_member(r[x][2],BITMAPINFOHEADER,"BitCount")<9 then
		r[x]=append(r[x],load_palette(fn,
				PE_get_member(r[x][1],ICONDIRENTRY,"ColorCount")))
	    else
		r[x]=append(r[x],{})
	    end if
		
	    -- get color pixels
	    r[x]=append(r[x],load_pixels(fn,
			      PE_get_member(r[x][1],ICONDIRENTRY,"Width"   ),
			      PE_get_member(r[x][1],ICONDIRENTRY,"Height"  ),
			      PE_get_member(r[x][2],BITMAPINFOHEADER,"BitCount")))

	    r[x]=append(r[x],load_pixels(fn,
			      PE_get_member(r[x][1],ICONDIRENTRY,"Width"   ),
			      PE_get_member(r[x][1],ICONDIRENTRY,"Height"  ),
			      1))
	end for
	return r
    else
	return-1
    end if
end function

function ico_set_start(integer fn,integer n)
    atom p
    p=where(fn)
    if seek(fn,PE_size_of(ICONDIR)+(n-1)*PE_size_of(ICONDIRENTRY)+12) then end if
    puts(fn,int_to_bytes(p))
    if seek(fn,p) then end if
    return p
end function

procedure ico_set_size(integer fn,integer n,integer start)
    atom p
    p=where(fn)
    if seek(fn,PE_size_of(ICONDIR)+(n-1)*PE_size_of(ICONDIRENTRY)+8) then end if
    puts(fn,int_to_bytes(p-start))
    if seek(fn,p) then end if
end procedure

function save_ico(sequence dat,sequence path)
    -- dat is { bpp,palette,XORbm,ANDbm }
    -- or a sequence of { bpp,palette,XORbm,ANDbm } objects
    sequence s
    integer fn,os
    if atom(dat[1]) then
	dat={dat}
    end if
    fn=open(path,"wb")
    if fn>1 then
	s=PE_blank_struct(ICONDIR)
	s=PE_set_member(s,ICONDIR,"Type" ,1)
	s=PE_set_member(s,ICONDIR,"Count",length(dat))
	PE_write_struct_to_file(fn,ICONDIR,s)
	for x=1 to length(dat) do
	    s=PE_blank_struct(ICONDIRENTRY)
	    s=PE_set_member(s,ICONDIRENTRY,"Width"        ,length(dat[x][3][1]))
	    s=PE_set_member(s,ICONDIRENTRY,"Height"       ,length(dat[x][3]   ))
	    s=PE_set_member(s,ICONDIRENTRY,"BitCount"     ,dat[x][1])
	    s=PE_set_member(s,ICONDIRENTRY,"ColorCount"   ,get_cc(dat[x][1]))
	    PE_write_struct_to_file(fn,ICONDIRENTRY,s)
	end for
	for x=1 to length(dat) do
	    s=PE_blank_struct(BITMAPINFOHEADER)
	    s=PE_set_member(s,BITMAPINFOHEADER,"Size"      ,40)
	    s=PE_set_member(s,BITMAPINFOHEADER,"Width"     ,length(dat[x][3][1]))
	    s=PE_set_member(s,BITMAPINFOHEADER,"Height"    ,length(dat[x][3])*2 )
	    s=PE_set_member(s,BITMAPINFOHEADER,"Planes"    ,1)
	    s=PE_set_member(s,BITMAPINFOHEADER,"BitCount"  ,dat[x][1])
	    os=ico_set_start(fn,x)
	    PE_write_struct_to_file(fn,BITMAPINFOHEADER,s)
	    if dat[x][1]<9 then
		write_palette(fn,dat[x][2])
	    end if
	    write_pixels(fn,dat[x][1],dat[x][3])
	    write_pixels(fn,1,dat[x][4])
	    ico_set_size(fn,x,os)
	end for
	return 0
    else
	return 1
    end if
end function


