-- Routines for working with PE files                             G.Burke 11/04
include PE_struc.e
include file.e

global constant  
    
    IMAGE_DOS_HEADER= {
		    {"e_magic"      ,2},
		    {"e_cblp"       ,2},
		    {"e_cp"         ,2},
		    {"e_crlc"       ,2},
		    {"e_cparhdr"    ,2},
		    {"e_minalloc"   ,2},
		    {"e_maxalloc"   ,2},
		    {"e_ss"         ,2},
		    {"e_sp"         ,2},
		    {"e_csum"       ,2},
		    {"e_ip"         ,2},
		    {"e_cs"         ,2},
		    {"e_lfarlc"     ,2},
		    {"e_ovno"       ,2},
		    {"e_res_1"      ,2},
		    {"e_res_2"      ,2},
		    {"e_res_3"      ,2},
		    {"e_res_4"      ,2},
		    {"e_oemid"      ,2},
		    {"e_oeminfo"    ,2},
		    {"e_res2_1"     ,2},
		    {"e_res2_2"     ,2},
		    {"e_res2_3"     ,2},
		    {"e_res2_4"     ,2},
		    {"e_res2_5"     ,2},
		    {"e_res2_6"     ,2},
		    {"e_res2_7"     ,2},
		    {"e_res2_8"     ,2},
		    {"e_res2_9"     ,2},
		    {"e_res2_A"     ,2},
		    {"e_lfanew"     ,4}     },
    
    
    IMAGE_SIGNATURE={   
			{"Signature",4}     },
    
    IMAGE_FILE_HEADER={
		    {"Machine"              ,2},
		    {"NumberOfSections"     ,2},
		    {"TimeDateStamp"        ,4},
		    {"PointerToSymbolTable" ,4},
		    {"NumberOfSymbols"      ,4},
		    {"SizeOfOptionalHeader" ,2},
		    {"Characteristics"      ,2}   },

    IMAGE_OPTIONAL_HEADER={
		    {"Magic"                        ,2},
		    {"MajorLinkerVersion"           ,1},
		    {"MinorLinkerVersion"           ,1},
		    {"SizeOfCode"                   ,4},
		    {"SizeOfInitializedData"        ,4},
		    {"SizeOfUninitializedData"      ,4},
		    {"AddressOfEntryPoint"          ,4},
		    {"BaseOfCode"                   ,4},
		    {"BaseOfData"                   ,4},
		      ------------------------------
		    {"ImageBase"                    ,4},
		    {"SectionAlignment"             ,4},
		    {"FileAlignment"                ,4},
		    {"MajorOperatingSystemVersion"  ,2},
		    {"MinorOperatingSystemVersion"  ,2},
		    {"MajorImageVersion"            ,2},
		    {"MinorImageVersion"            ,2},
		    {"MajorSubsystemVersion"        ,2},
		    {"MinorSubsystemVersion"        ,2},
		    {"Reserved1"                    ,4},
		    {"SizeOfImage"                  ,4},
		    {"SizeOfHeaders"                ,4},
		    {"CheckSum"                     ,4},
		    {"Subsystem"                    ,2},
		    {"DllCharacteristics"           ,2},
		    {"SizeOfStackReserve"           ,4},
		    {"SizeOfStackCommit"            ,4},
		    {"SizeOfHeapReserve"            ,4},
		    {"SizeOfHeapCommit"             ,4},
		    {"LoaderFlags"                  ,4},
		    {"NumberOfRvaAndSizes"          ,4}     },


    IMAGE_NUMBEROF_DIRECTIORY_ENTRIES=#10,

    IMAGE_DATA_DIRECTORY={
		    {"VirtualAddress",4},
		    {"Size",4}              },
    IDD_NAMES={
		    "EXPORT",
		    "IMPORT",
		    "RESOURCE",
		    "EXCEMPTION",
		    "SECURITY",
		    "BASERELOC",
		    "DEBUG",
		    "COPYRIGHT",
		    "GLOBALPTR",
		    "TLS",
		    "LOAD CONFIG",
		    "BOUND IMPORT",
		    "IMPORT ADDRESS TABLE",
		    "IDD 14",
		    "IDD 15",
		    "IDD 16"      },
    
    IMAGE_SIZEOF_SHORT_NAME=8,

    IMAGE_SECTION_HEADER={
		    {"Name",IMAGE_SIZEOF_SHORT_NAME},
		    {"VirtualSize"          ,4},
		    {"VirtualAddress"       ,4},
		    {"SizeOfRawData"        ,4},
		    {"PointerToRawData"     ,4},
		    {"PointerToRelocations" ,4},
		    {"PointerToLinenumbers" ,4},
		    {"NumberOfRelocations"  ,2},
		    {"NumberOfLinenumbers"  ,2},
		    {"Characteristics"      ,4}     },

    IMAGE_RELOCATION={
		    {"VirtualAddress"       ,4},
		    {"SymbolTableIndex"     ,4},
		    {"Type"                 ,2}     },
    
    IMAGE_DOS_SIGNATURE    =#5A4D,          --MZ
    IMAGE_NT_SIGNATURE     =#00004550,      --PE\0\0

    STUB_PROG={     #0E,#1F,#BA,#0E,#00,#B4,#09,
		    #CD,#21,#B8,#01,#4C,#CD,#21     },

    IMAGE_FILE_MACHINE_I386=#14C,

    IMAGE_SUBSYSTEM_WINDOWS_GUI=2,
    IMAGE_SUBSYSTEM_WINDOWS_CUI=3,
    
    ------------------Section Characteristics--------------------

--  IMAGE_SCN_TYPE_REG                   =#00000000,  -- Reserved.
--  IMAGE_SCN_TYPE_DSECT                 =#00000001,  -- Reserved.
--  IMAGE_SCN_TYPE_NOLOAD                =#00000002,  -- Reserved.
--  IMAGE_SCN_TYPE_GROUP                 =#00000004,  -- Reserved.
--  IMAGE_SCN_TYPE_NO_PAD                =#00000008,  -- Reserved.
--  IMAGE_SCN_TYPE_COPY                  =#00000010,  -- Reserved.
    IMAGE_SCN_CNT_CODE                   =#00000020,  -- contains code
    IMAGE_SCN_CNT_INITIALIZED_DATA       =#00000040,  -- contains initialized data
    IMAGE_SCN_CNT_UNINITIALIZED_DATA     =#00000080,  -- contains uninitialized data.
--  IMAGE_SCN_LNK_OTHER                  =#00000100,  -- Reserved.
--  IMAGE_SCN_LNK_INFO                   =#00000200,  -- contains comments or some other type of information.
--  IMAGE_SCN_TYPE_OVER                  =#00000400,  -- Reserved.
--  IMAGE_SCN_LNK_REMOVE                 =#00000800,  -- contents will not become part of image.
--  IMAGE_SCN_LNK_COMDAT                 =#00001000,  -- contents comdat.
--  IMAGE_SCN_MEM_PROTECTED - Obsolete   =#00004000,
--  IMAGE_SCN_MEM_FARDATA                =#00008000,
--  IMAGE_SCN_MEM_SYSHEAP  - Obsolete    =#00010000,
--  IMAGE_SCN_MEM_PURGEABLE              =#00020000,
--  IMAGE_SCN_MEM_16BIT                  =#00020000,
--  IMAGE_SCN_MEM_LOCKED                 =#00040000,
--  IMAGE_SCN_MEM_PRELOAD                =#00080000,
--  IMAGE_SCN_ALIGN_1BYTES               =#00100000,
--  IMAGE_SCN_ALIGN_2BYTES               =#00200000,
--  IMAGE_SCN_ALIGN_4BYTES               =#00300000,
--  IMAGE_SCN_ALIGN_8BYTES               =#00400000,
--  IMAGE_SCN_ALIGN_16BYTES              =#00500000,  -- Default alignment if no others are specified.
--  IMAGE_SCN_ALIGN_32BYTES              =#00600000,
--  IMAGE_SCN_ALIGN_64BYTES              =#00700000,
--  Unused                               =#00800000,
--  IMAGE_SCN_LNK_NRELOC_OVFL            =#01000000,  -- contains extended relocations.
    IMAGE_SCN_MEM_DISCARDABLE            =#02000000,  -- can be discarded.
    IMAGE_SCN_MEM_NOT_CACHED             =#04000000,  -- is not cachable.
    IMAGE_SCN_MEM_NOT_PAGED              =#08000000,  -- is not pageable.
    IMAGE_SCN_MEM_SHARED                 =#10000000,  -- is shareable.
    IMAGE_SCN_MEM_EXECUTE                =#20000000,  -- is executable.
    IMAGE_SCN_MEM_READ                   =#40000000,  -- is readable.
    IMAGE_SCN_MEM_WRITE                  =#80000000,  -- is writeable.
    
    -- hdr_info pseudo structure
    hi_TimeDateStamp                =1,
    hi_Characteristics              =2,
    hi_SectionAlignment             =3,
    hi_FileAlignment                =4,
    hi_MajorOperatingSystemVersion  =5,
    hi_MinorOperatingSystemVersion  =6,
    hi_MajorImageVersion            =7,
    hi_MinorImageVersion            =8,
    hi_MajorSubsystemVersion        =9,
    hi_MinorSubsystemVersion        =10,
    hi_Subsystem                    =11,
    hi_SizeOfStackReserve           =12,
    hi_SizeOfStackCommit            =13,
    hi_SizeOfHeapReserve            =14,
    hi_SizeOfHeapCommit             =15,
    hi_EntryPointOffset             =16, -- os into first executable section
    hi_ImageBase                    =17,
    hi_StubMessage                  =18,

    -- section header pseudo structure
    sh_Name             =1,
    sh_Characteristics  =2,
    sh_DirEntry         =3,
    sh_Data             =4, -- actual data or {path,offset,size} for external
    -- fields added by system
    sh_FileOffset       =5,
    sh_RVA              =6,

    IMAGE_EXPORT_DIRECTORY= {
		    {"Characteristics"      ,4},
		    {"TimeDateStamp"        ,4},
		    {"MajorVersion"         ,2},
		    {"MinorVersion"         ,2},
		    {"Name"                 ,4},
		    {"Base"                 ,4},
		    {"NumberOfFunctions"    ,4},
		    {"NumberOfNames"        ,4},
		    {"AddressOfFunctions"   ,4},
		    {"AddressOfNames"       ,4},
		    {"AddressOfNameOrdinals",4}
						    },

    IMAGE_RESOURCE_DIRECTORY=   {
		    {"Characteristics"      ,4},
		    {"TimeDateStamp"        ,4},
		    {"MajorVersion"         ,2},
		    {"MinorVersion"         ,2},
		    {"NumberOfNamedEntries" ,2},
		    {"NumberOfIdEntries"    ,2}     },
    
    IMAGE_RESOURCE_NAME_IS_STRING   =#80000000,
    IMAGE_RESOURCE_DATA_IS_DIRECTORY=#80000000,
    
    IMAGE_RESOURCE_DATA_ENTRY=  {
			
		    {"OffsetToData"         ,4},
		    {"Size"                 ,4},
		    {"CodePage"             ,4},
		    {"Reserved"             ,4}     },
    
    RSRC_TYPES= {   
		    {"UNKNOWN"              ,#00},
		    {"CURSOR"               ,#01},
		    {"BITMAP"               ,#02},
		    {"ICON"                 ,#03},
		    {"MENU"                 ,#04},
		    {"DIALOG"               ,#05},
		    {"STRING"               ,#06},
		    {"FONTDIR"              ,#07},
		    {"FONT"                 ,#08},
		    {"ACCELERATOR"          ,#09},
		    {"RCDATA"               ,#0A},
		    {"MESSAGETABLE"         ,#0B},
		    {"GROUP_CURSOR"         ,#0C},
		    {"GROUP_ICON"           ,#0E},
		    {"VERSION"              ,#10},
		    {"MAINFEST"             ,#18},
		    {"DIALOG DATA"          ,#F0},  
		    {"TOOLBAR"              ,#F1},   
		    {"DIRECTORY"            ,#1000}   },
    
		    RSRC_UNICODE=1,    --(true)
		    
	    -- Resource pseudo structure field defs
		    
		    RSRC_Name           =01,
		    RSRC_Characteristics=02,
		    RSRC_TimeDateStamp  =03,
		    RSRC_MajorVersion   =04,
		    RSRC_MinorVersion   =05,
		    RSRC_NumberOfEntries=06,
		    RSRC_Type           =07,
		    RSRC_CodePage       =08,
		    RSRC_Reserved       =09,
		    RSRC_Object         =10,
		    RSRC_Parent         =11,
		    RSRC_EuObject       =12,
    
		    number_of_rsrc_fields =12,


	hdr_info_defaults={    120870,
				#8182,
				#1000,
				#200,
				1,
				#B,
				0,
				0,
				3,
				#A,
				2,
				#10000,
				#10000,
				#2000,
				#1000,
				0,
				#400000,
			"This Euphoria Program Requires Windows"&#D&#A    },

	 section_defaults={ --    name  characteristics DirEntry 
				{".text"  ,#60000020    ,0},
				{"BEGTEXT",#60000020    ,0},
				{"DGROUP" ,#C0000040    ,0},
				{".bss"   ,#C0000080    ,0},
				{".idata" ,#C0000040    ,2},
				{".reloc" ,#42000040    ,6},
				{"CODE"   ,#E0000020    ,0},
				{"DATA"   ,#C0000040    ,0},
				{".rsrc"  ,#CC000040    ,3}     },




---------------------------------------------------------------------------
--          Language macros  -  from winNT.h
---------------------------------------------------------------------------

--global function MAKELANGID(integer p,integer s)
    -- returns language id from a primary language id and a sublanguage id.

--global function PRIMARYLANGID(integer lgid)
    -- returns primary language id from a language id.

--global function SUBLANGID(integer lgid)
    -- returns sublanguage id from a language id.

--global function MAKELCID(integer lgid,integer srtid)
    -- returns locale ID from language ID and sort ID

--global function LANGIDFROMLCID(atom lcid)
    -- returns language ID from locale ID

--global function SORTIDFROMLCID(atom lcid)
    -- returns sort ID from locale ID


---------------------------------------------------------------------------
---------------------------------------------------------------------------
--  *
--  *  Language IDs.
--  *
--  *  The following two combinations of primary language ID and
--  *  sublanguage ID have special semantics:
--  *
--  *    Primary Language ID   Sublanguage ID      Result
--  *    -------------------   ---------------     ------------------------
--  *    LANG_NEUTRAL          SUBLANG_NEUTRAL     Language neutral
--  *    LANG_NEUTRAL          SUBLANG_DEFAULT     User default language
--  *    LANG_NEUTRAL          SUBLANG_SYS_DEFAULT System default language

--  *
--  *  Primary language IDs.
--  *

	LANG_NEUTRAL        =#00,
--      LANG_ARABIC         =#01,
--      LANG_AFRIKAANS      =#36,
--      LANG_ALBANIAN       =#1C,
--      LANG_BASQUE         =#2D,
--      LANG_BULGARIAN      =#02,
--      LANG_BYELORUSSIAN   =#23,
--      LANG_CATALAN        =#03,
--      LANG_CHINESE        =#04,
--      LANG_CROATIAN       =#1A,
--      LANG_CZECH          =#05,
--      LANG_DANISH         =#06,
--      LANG_DUTCH          =#13,
--      LANG_ENGLISH        =#09,
--      LANG_ESTONIAN       =#25,
--      LANG_FAEROESE       =#38,
--      LANG_FARSI          =#29,
--      LANG_FINNISH        =#0B,
--      LANG_FRENCH         =#0C,
--      LANG_GERMAN         =#07,
--      LANG_GREEK          =#08,
--      LANG_HEBREW         =#0D,
--      LANG_HUNGARIAN      =#0E,
--      LANG_ICELANDIC      =#0F,
--      LANG_INDONESIAN     =#21,
--      LANG_ITALIAN        =#10,
--      LANG_JAPANESE       =#11,
--      LANG_KOREAN         =#12,
--      LANG_LATVIAN        =#26,
--      LANG_LITHUANIAN     =#27,
--      LANG_NORWEGIAN      =#14,
--      LANG_POLISH         =#15,
--      LANG_PORTUGUESE     =#16,
--      LANG_ROMANIAN       =#18,
--      LANG_RUSSIAN        =#19,
--      LANG_SLOVAK         =#1B,
--      LANG_SLOVENIAN      =#24,
--      LANG_SORBIAN        =#2E,
--      LANG_SPANISH        =#0A,
--      LANG_SWEDISH        =#1D,
--      LANG_THAI           =#1E,
--      LANG_TURKISH        =#1F,
--      LANG_UKRAINIAN      =#22,
-- 
--      LANG_NAMES={
--              "NEUTRAL",
--              "ARABIC",
--              "BULGARIAN",
--              "CATALAN",
--              "CHINESE",
--              "CZECH ",
--              "DANISH",
--              "GERMAN",
--              "GREEK",
--              "ENGLISH",
--              "SPANISH",
--              "FINNISH",
--              "FRENCH",
--              "HEBREW",
--              "HUNGARIAN",
--              "ICELANDIC",
--              "ITALIAN",
--              "JAPANESE",
--              "KOREAN",
--              "DUTCH",
--              "NORWEGIAN",
--              "POLISH",
--              "PORTUGUESE",
--              "",
--              "ROMANIAN",
--              "RUSSIAN",
--              "CROATIAN",
--              "SLOVAK",
--              "ALBANIAN",
--              "SWEDISH",
--              "THAI",
--              "TURKISH",
--              "",
--              "INDONESIAN",
--              "UKRAINIAN",
--              "BYELORUSSIAN",
--              "SLOVENIAN",
--              "ESTONIAN",
--              "LATVIAN",
--              "LITHUANIAN",
--              "",
--              "FARSI",
--              "","","",
--              "BASQUE",
--              "SORBIAN",
--              "","","","","","","",
--              "AFRIKAANS",
--              "",
--              "FAEROESE"
--                                      },
-- 
-- -- /*
-- --  *  Sublanguage IDs.
-- --  *
-- --  *  The name immediately following SUBLANG_ dictates which primary
-- --  *  language ID that sublanguage ID can be combined with to form a
-- --  *  valid language ID.
-- --  */
-- 
	SUBLANG_NEUTRAL                 =#00,
	SUBLANG_DEFAULT                 =#01,
	SUBLANG_SYS_DEFAULT             =#02,
-- 
--      SUBLANG_ARABIC                  =#01,--( SAUDI )
--      SUBLANG_ARABIC_IRAQ             =#02,
--      SUBLANG_ARABIC_EGYPT            =#03,
--      SUBLANG_ARABIC_LIBYA            =#04,
--      SUBLANG_ARABIC_ALGERIA          =#05,
--      SUBLANG_ARABIC_MOROCCO          =#06,
--      SUBLANG_ARABIC_TUNISIA          =#07,
--      SUBLANG_ARABIC_OMAN             =#08,
--      SUBLANG_ARABIC_YEMEN            =#09,
--      SUBLANG_ARABIC_SYRIA            =#10,
--      SUBLANG_ARABIC_JORDAN           =#11,
--      SUBLANG_ARABIC_LEBANON          =#12,
--      SUBLANG_ARABIC_KUWAIT           =#13,
--      SUBLANG_ARABIC_UAE              =#14,
--      SUBLANG_ARABIC_BAHRAIN          =#15,
--      SUBLANG_ARABIC_QATAR            =#16,
-- 
--      SUBLANG_CHINESE_TRADITIONAL     =#01,
--      SUBLANG_CHINESE_SIMPLIFIED      =#02,
--      SUBLANG_CHINESE_HONGKONG        =#03,
--      SUBLANG_CHINESE_SINGAPORE       =#04,
-- 
--      SUBLANG_DUTCH                   =#01,
--      SUBLANG_DUTCH_BELGIAN           =#02,
-- 
--      SUBLANG_ENGLISH_US              =#01,
--      SUBLANG_ENGLISH_UK              =#02,
--      SUBLANG_ENGLISH_AUS             =#03,
--      SUBLANG_ENGLISH_CAN             =#04,
--      SUBLANG_ENGLISH_NZ              =#05,
--      SUBLANG_ENGLISH_EIRE            =#06,
--      SUBLANG_ENGLISH_SAFRICA         =#07,
--      SUBLANG_ENGLISH_JAMAICA         =#08,
--      SUBLANG_ENGLISH_CARRIBEAN       =#09,
-- 
--      SUBLANG_FRENCH                  =#01,
--      SUBLANG_FRENCH_BELGIAN          =#02,
--      SUBLANG_FRENCH_CANADIAN         =#03,
--      SUBLANG_FRENCH_SWISS            =#04,
--      SUBLANG_FRENCH_LUXEMBOURG       =#05,
-- 
--      SUBLANG_GERMAN                  =#01,
--      SUBLANG_GERMAN_SWISS            =#02,
--      SUBLANG_GERMAN_AUSTRIAN         =#03,
--      SUBLANG_GERMAN_LUXEMBOURG       =#04,
--      SUBLANG_GERMAN_LIECHTENSTEIN    =#05,
-- 
--      SUBLANG_ITALIAN                 =#01,
--      SUBLANG_ITALIAN_SWISS           =#02,
-- 
--      SUBLANG_KOREAN                  =#01,
--      SUBLANG_KOREAN_JOHAB            =#02,
-- 
--      SUBLANG_NORWEGIAN_BOKMAL        =#01,
--      SUBLANG_NORWEGIAN_NYNORSK       =#02,
-- 
--      SUBLANG_PORTUGUESE              =#02,
--      SUBLANG_PORTUGUESE_BRAZILIAN    =#01,
-- 
--      SUBLANG_SPANISH                 =#01,
--      SUBLANG_SPANISH_MEXICAN         =#02,
--      SUBLANG_SPANISH_MODERN          =#03,
--      SUBLANG_SPANISH_GUATEMALA       =#04,
--      SUBLANG_SPANISH_COSTARICA       =#05,
--      SUBLANG_SPANISH_PANAMA          =#06,
--      SUBLANG_SPANISH_DOMINICAN       =#07,
--      SUBLANG_SPANISH_VENEZUELA       =#08,
--      SUBLANG_SPANISH_COLOMBIA        =#09,
--      SUBLANG_SPANISH_PERU            =#10,
--      SUBLANG_SPANISH_ARGENTINA       =#11,
--      SUBLANG_SPANISH_ECUADOR         =#12,
--      SUBLANG_SPANISH_CHILE           =#13,
--      SUBLANG_SPANISH_URUGUAY         =#14,
--      SUBLANG_SPANISH_PARAGUAY        =#15,
--      SUBLANG_SPANISH_BOLIVIA         =#16,
-- 
-- -- Zero based i.e.  
-- --      s=SUBLANG_NAMES[LANG_ARABIC+1][SUBLANG_ARABIC_IRAQ+1]
-- -- MyName=SUBLANG_NAMES[LANG_HUNGARIAN+1][SUBLANG_NEUTRAL+1]
--      
--      SUBLANG_NAMES=
-- 
--          {{  "NEUTRAL",
--              "DEFAULT",
--              "SYS_DEFAULT"},
-- 
--          {"","ARABIC_SAUDI",
--              "ARABIC_IRAQ",
--              "ARABIC_EGYPT",
--              "ARABIC_LIBYA",
--              "ARABIC_ALGERIA",
--              "ARABIC_MOROCCO",
--              "ARABIC_TUNISIA",
--              "ARABIC_OMAN",
--              "ARABIC_YEMEN","","","","","","",
--              "ARABIC_SYRIA",
--              "ARABIC_JORDAN",
--              "ARABIC_LEBANON",
--              "ARABIC_KUWAIT",
--              "ARABIC_UAE",
--              "ARABIC_BAHRAIN",
--              "ARABIC_QATAR"},{},{},
--          
--          {"","CHINESE_TRADITIONAL",
--              "CHINESE_SIMPLIFIED",
--              "CHINESE_HONGKONG",
--              "CHINESE_SINGAPORE"},{},{},
--              
--          {"","GERMAN",
--              "GERMAN_SWISS",
--              "GERMAN_AUSTRIAN",
--              "GERMAN_LUXEMBOURG",
--              "GERMAN_LIECHTENSTEIN"},{},
--          
--          {"","ENGLISH_US",
--              "ENGLISH_UK",
--              "ENGLISH_AUS",
--              "ENGLISH_CAN",
--              "ENGLISH_NZ",
--              "ENGLISH_EIRE",
--              "ENGLISH_SAFRICA",
--              "ENGLISH_JAMAICA",
--              "ENGLISH_CARRIBEAN"},
--          
--          {"","SPANISH",
--              "SPANISH_MEXICAN",
--              "SPANISH_MODERN",
--              "SPANISH_GUATEMALA",
--              "SPANISH_COSTARICA",
--              "SPANISH_PANAMA",
--              "SPANISH_DOMINICAN",
--              "SPANISH_VENEZUELA",
--              "SPANISH_COLOMBIA","","","","","","",
--              "SPANISH_PERU",
--              "SPANISH_ARGENTINA",
--              "SPANISH_ECUADOR",
--              "SPANISH_CHILE",
--              "SPANISH_URUGUAY",
--              "SPANISH_PARAGUAY",
--              "SPANISH_BOLIVIA"},{},
--          
--          {"","FRENCH",
--              "FRENCH_BELGIAN",
--              "FRENCH_CANADIAN",
--              "FRENCH_SWISS",
--              "FRENCH_LUXEMBOURG"},{},{},{},
--          
--          {"","ITALIAN",
--              "ITALIAN_SWISS"},{},
--          
--          {"","KOREAN",
--              "KOREAN_JOHAB"},
--          
--          {"","DUTCH",
--              "DUTCH_BELGIAN"},
--          
--          {"","NORWEGIAN_BOKMAL",
--              "NORWEGIAN_NYNORSK"},{},
--          
--          {"","PORTUGUESE",
--              "PORTUGUESE_BRAZILIAN"}},
--      
--      
--      
-- -- /*
-- --  *  Sorting IDs.
-- --  *
-- --  */
--  
    SORT_DEFAULT                     =0--,     --/* sorting default */
-- 
--  SORT_JAPANESE_XJIS               =0,     --/* Japanese XJIS order */
--  SORT_JAPANESE_UNICODE            =1,     --/* Japanese Unicode order */
-- 
--  SORT_CHINESE_BIG5                =0,     --/* Chinese BIG5 order */
--  SORT_CHINESE_UNICODE             =1,     --/* Chinese Unicode order */
-- 
--  SORT_KOREAN_KSC                  =0,     --/* Korean KSC order */
--  SORT_KOREAN_UNICODE              =1      --/* Korean Unicode order */
-- 
-- -- /*->from winnt.h
-- --  *
-- --  *  A language ID is a 16 bit value which is the combination of a
-- --  *  primary language ID and a secondary language ID.  The bits are
-- --  *  allocated as follows:
-- --  *
-- --  *       +-----------------------+-------------------------+
-- --  *       |     Sublanguage ID    |   Primary Language ID   |
-- --  *       +-----------------------+-------------------------+
-- --  *        15                   10 9                       0   bit
-- --  *
-- --  *
-- --  *  Language ID creation/extraction macros:
-- --  *
-- --  *    MAKELANGID    - construct language id from a primary language id and
-- --  *                    a sublanguage id.
-- --  *    PRIMARYLANGID - extract primary language id from a language id.
-- --  *    SUBLANGID     - extract sublanguage id from a language id.
-- --  */
-- -- #define MAKELANGID(p, s)       ((((WORD  )(s)) << 10) | (WORD  )(p))
-- -- #define PRIMARYLANGID(lgid)    ((WORD  )(lgid) & 0x3ff)
-- -- #define SUBLANGID(lgid)        ((WORD  )(lgid) >> 10)

global function MAKELANGID(integer p,integer s)
    return s*#400+p
end function

global function PRIMARYLANGID(integer lgid)
    return and_bits(lgid,#3FF)
end function

global function SUBLANGID(integer lgid)
    return floor(lgid/#400)
end function

-- /*
--  *  A locale ID is a 32 bit value which is the combination of a
--  *  language ID, a sort ID, and a reserved area.  The bits are
--  *  allocated as follows:
--  *
--  *       +-------------+---------+-------------------------+
--  *       |   Reserved  | Sort ID |      Language ID        |
--  *       +-------------+---------+-------------------------+
--  *        31         20 19     16 15                      0   bit
--  *
--  *
--  *  Locale ID creation/extraction macros:
--  *
--  *    MAKELCID       - construct locale id from a language id and a sort id.
--  *    LANGIDFROMLCID - extract language id from a locale id.
--  *    SORTIDFROMLCID - extract sort id from a locale id.
--  */

constant NLS_VALID_LOCALE_MASK  =#000FFFFF

-- #define MAKELCID(lgid, srtid)  ((DWORD)((((DWORD)((WORD  )(srtid))) << 16) |  \
--                                       ((DWORD)((WORD  )(lgid)))))
-- #define LANGIDFROMLCID(lcid)   ((WORD  )(lcid))
-- #define SORTIDFROMLCID(lcid)   ((WORD  )((((DWORD)(lcid)) & NLS_VALID_LOCALE_MASK) >> 16))

global function MAKELCID(integer lgid,integer srtid)
    return srtid*#10000+lgid
end function

global function LANGIDFROMLCID(atom lcid)
    return and_bits(lcid,#FFFF)
end function

global function SORTIDFROMLCID(atom lcid)
    return floor(and_bits(lcid,NLS_VALID_LOCALE_MASK)/#10000)
end function

-- /*
--  *  Default System and User IDs for language and locale.
--  */

constant
    LANG_SYSTEM_DEFAULT    =MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT),
    LANG_USER_DEFAULT      =MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
    LOCALE_SYSTEM_DEFAULT  =MAKELCID(LANG_SYSTEM_DEFAULT, SORT_DEFAULT),
    LOCALE_USER_DEFAULT    =MAKELCID(LANG_USER_DEFAULT, SORT_DEFAULT)
    
    
if atom(LOCALE_SYSTEM_DEFAULT&LOCALE_USER_DEFAULT) then end if



-------------------------------------------------------------------------------
-------------------<PE file structure handler>---------------------------------
-------------------------------------------------------------------------------


global function PE_load_header(sequence path)
    --returns {     dos header,
    --              im_file header,
    --              im_opt header,
    --              {im_data directories},
    --              {im_sect headers}         }
    sequence r,l
    integer fn
    
    r=repeat(0,5)
    fn=open(path,"rb")
    if fn<2 then
	PE_last_error("PE_load_header : couldnt open "&path)
	return 1
    end if
    r[1]=PE_get_struct_from_file(fn,IMAGE_DOS_HEADER)
    if compare(PE_get_member(r[1],IMAGE_DOS_HEADER,"e_magic"),
						    IMAGE_DOS_SIGNATURE) then
	PE_last_error("PE_load_header : DOS header not found")
	return 2
    end if
    if seek(fn,PE_get_member(r[1],IMAGE_DOS_HEADER,"e_lfanew")) then end if
    if compare(PE_get_struct_from_file(fn,IMAGE_SIGNATURE),{IMAGE_NT_SIGNATURE}) then
	PE_last_error("PE_load_header : PE header not found ( DOS Image )")
	return 3
    end if
    r[2]=PE_get_struct_from_file(fn,IMAGE_FILE_HEADER)
    r[3]=PE_get_struct_from_file(fn,IMAGE_OPTIONAL_HEADER)
    l={}
    for x=1 to PE_get_member(r[3],IMAGE_OPTIONAL_HEADER,"NumberOfRvaAndSizes") do
	l=append(l,PE_get_struct_from_file(fn,IMAGE_DATA_DIRECTORY))
    end for
    r[4]=l
    l={}
    for x=1 to PE_get_member(r[2],IMAGE_FILE_HEADER,"NumberOfSections") do
	l=append(l,PE_get_struct_from_file(fn,IMAGE_SECTION_HEADER))
    end for
    r[5]=l
    close(fn)
    return r
end function

global function PE_lookup_section(sequence pe,sequence name)
    -- returns index of named section or 0 if not present
    sequence s
    
    s=repeat(0,IMAGE_SIZEOF_SHORT_NAME)
    for x=1 to length(name) do
	if x<=length(s) then
	    s[x]=name[x]
	end if
    end for
    for x=1 to PE_get_member(pe[2],IMAGE_FILE_HEADER,"NumberOfSections")  do
	if equal(PE_get_member(pe[5][x],IMAGE_SECTION_HEADER,"Name"),s) then
	    return x
	end if
    end for
    PE_last_error("PE_lookup_section : section not found")
    return 0
end function

function find_section_loc(sequence pe,object section)
    --returns {offset,size}
    integer l
    atom os
    
    if sequence(section) then
	section=PE_lookup_section(pe,section)
    end if
    if section then
	os=PE_get_member(pe[5][section],IMAGE_SECTION_HEADER,"PointerToRawData")
	if not os then return {0,0} end if
	l=PE_get_member(pe[5][section],IMAGE_SECTION_HEADER,"VirtualSize")
	if not l then
	    l=PE_get_member(pe[5][section],IMAGE_SECTION_HEADER,"SizeOfRawData")
	end if
	-- get actual size for sections with dir entries
	for x=1 to length(pe[4]) do
	    if PE_get_member(pe[4][x],IMAGE_DATA_DIRECTORY,"VirtualAddress")=
       PE_get_member(pe[5][section],IMAGE_SECTION_HEADER,"VirtualAddress") then
		if PE_get_member(pe[4][x],IMAGE_DATA_DIRECTORY,"Size") then
		    l=PE_get_member(pe[4][x],IMAGE_DATA_DIRECTORY,"Size")
		    exit
		end if
	    end if
	end for
	return {os,l}
    end if
    PE_last_error("find_section_loc : section not found")
    return 0
end function

global function PE_get_section(sequence path,sequence pe,object section)
    --returns raw section data as sequence of bytes
    integer fn
    sequence r,loc
    
    loc=find_section_loc(pe,section)
    if sequence(loc) then
	fn=open(path,"rb")
	if fn>1 then
	    if seek(fn,loc[1]) then end if
	    r=repeat(0,loc[2])
	    for x=1 to loc[2] do
		r[x]=getc(fn)
	    end for
	    close(fn)
	    return r
	end if
	PE_last_error("PE_get_section : cant open "&path)
	return 0
    end if
    PE_last_error("PE_get_section : section not found")
    return 0
end function

				
function get_info_defs(object hdr_info)
    if sequence(hdr_info) then
	for x=1 to length(hdr_info) do
	    if equal(hdr_info[x],0) then
		hdr_info[x]=hdr_info_defaults[x]
	    end if
	end for
	return hdr_info
    else
	return hdr_info_defaults
    end if
end function

procedure PE_write_to(integer fn,atom os)
    -- write zeros to file fn until offset os is reached
    puts(fn,repeat(0,os-where(fn)))
end procedure

global function PE_round_to(atom os,integer border)
    -- round os up to the next border byte boundary
    if remainder(os,border) then
	return floor(os/border)*border+border
    else
	return os
    end if
end function

function get_sec_defs(sequence name)
    for x=1 to length(section_defaults)  do
	if equal(section_defaults[x][1],name) then
	    return section_defaults[x][2..3]
	end if
    end for
    return {0,0}
end function

function get_section_vsize(sequence s)
    if length(s[sh_Data])=3 then
	return s[sh_Data][3]
    else
	return length(s[sh_Data])
    end if
end function

function get_section_size(sequence s)
    if length(s[sh_Data])=3 then
	if compare(s[sh_Data][1],{}) or compare(s[sh_Data][2],0)  then
	    return s[sh_Data][3]
	else
	    return 0
	end if
    else
	return length(s[sh_Data])
    end if
end function

function calc_section_offsets(sequence sec,sequence hdr_info)
    atom file,mem
    file=0 mem=hdr_info[hi_SectionAlignment]
    for y=1 to length(sec) do
	for x=1 to length(sec[y]) do
	    sec[y][x][sh_FileOffset]=file
	    sec[y][x][sh_RVA]       =mem
	    file+= PE_round_to(get_section_size(sec[y][x]),hdr_info[hi_FileAlignment])
	    mem += PE_round_to(get_section_vsize(sec[y][x]),hdr_info[hi_SectionAlignment])
	end for
    end for
    return sec
end function

function process_sections(sequence sec,sequence hdr_info)
    sequence exec,code  ,udata,idata,
	     rsrc,import,other,reloc
    
    exec={}   code={} udata={} idata={} 
    rsrc={} import={} other={} reloc={}
    
    -- get defaults if not set
    for x=1 to length(sec) do
	if equal(sec[x][2..3],{0,0}) then
	    sec[x][2..3]=get_sec_defs(sec[x][1])
	end if              
	sec[x]&=0&0
	if    and_bits(sec[x][sh_Characteristics],IMAGE_SCN_MEM_EXECUTE) then
	    exec=append(exec,sec[x])
	elsif sec[x][sh_DirEntry]=2 then
	    import=append(import,sec[x])
	elsif sec[x][sh_DirEntry]=3 then
	    rsrc=append(rsrc,sec[x])
	elsif sec[x][sh_DirEntry]=6 then
	    reloc=append(reloc,sec[x])
	elsif and_bits(sec[x][sh_Characteristics],IMAGE_SCN_CNT_CODE) then
	    code=append(code,sec[x])
	elsif and_bits(sec[x][sh_Characteristics],IMAGE_SCN_CNT_UNINITIALIZED_DATA) then
	    udata=append(udata,sec[x])
	elsif and_bits(sec[x][sh_Characteristics],IMAGE_SCN_CNT_INITIALIZED_DATA) then
	    idata=append(idata,sec[x])
	else
	    other=append(other,sec[x])          
	end if
    end for
    -- order sections
    sec={   exec,  code, idata,udata,
	    import,reloc,rsrc, other    }
    return calc_section_offsets(sec,hdr_info)
end function

function get_code_size(sequence sec,sequence hdr_info)
    atom r
    r=0
    for y=1 to 2 do
	for x=1 to length(sec[y]) do
	    r+=PE_round_to(get_section_size(sec[y][x]),hdr_info[hi_FileAlignment])
	end for
    end for
    return r
end function

function get_init_size(sequence sec,sequence hdr_info)
    atom r
    r=0
    for y=3 to length(sec) do
	for x=1 to length(sec[y]) do
	    if and_bits(sec[y][x][2],IMAGE_SCN_CNT_INITIALIZED_DATA) then
		r+=PE_round_to(get_section_size(sec[y][x]),hdr_info[hi_FileAlignment])
	    end if
	end for
    end for
    return r
end function

function get_uninit_size(sequence sec,sequence hdr_info)
    atom r
    r=0
    for y=3 to length(sec) do
	for x=1 to length(sec[y]) do
	    if and_bits(sec[y][x][2],IMAGE_SCN_CNT_UNINITIALIZED_DATA) then
		r+=PE_round_to(get_section_size(sec[y][x]),hdr_info[hi_FileAlignment])
	    end if
	end for
    end for
    return r
end function

function get_image_size(sequence sec,sequence hdr_info)
    atom a,s
    for y=1 to length(sec) do
	if length(sec[y]) then
	    a=sec[y][length(sec[y])][6]
	    s=get_section_size(sec[y][length(sec[y])])
	end if
    end for
    return a+PE_round_to(s,hdr_info[hi_SectionAlignment])
end function

function get_hdr_size(sequence sec)
    return sec[1][1][sh_FileOffset]
end function

function get_dir_links(sequence sec)
    sequence r
    r=repeat(0,IMAGE_NUMBEROF_DIRECTIORY_ENTRIES)
    for y=1 to length(sec) do
	for x=1 to length(sec[y]) do
	    if sec[y][x][sh_DirEntry] then
		r[sec[y][x][sh_DirEntry]]=sec[y][x]
	    end if
	end for
    end for
    return r
end function

function write_section(integer fn,sequence sec)
    integer fn2
    if length(sec[sh_Data])=3 then
	if equal(sec[sh_Data][1],{}) then
	    return 0 --send {"",0,vsize}  for uninitialized data
		     --send {"",1,vsize}  for initialized data not written now
	end if
	fn2=open(sec[sh_Data][1],"rb")
	if fn2<2 then 
	    PE_last_error("write_section : cant open "&sec[sh_Data][1])
	    return 1 
	end if
	if seek(fn2,sec[sh_Data][2]) then   end if
	for x=1 to sec[sh_Data][3] do
	    puts(fn,getc(fn2))
	end for
	close(fn2)
    else
	puts(fn,sec[sh_Data])
    end if  
    return 0
end function

function get_data_base(sequence sec)
    for x=3 to length(sec) do
	if length(sec[x]) then
	    return sec[x][1][sh_RVA]
	end if
    end for
    PE_last_error("get_data_base : no data sections found")
    return 0
end function

global function write_PE_ex(    sequence path,
				object   hdr_info,
				sequence sections       )
    integer fn,os,nsect
    sequence s,t
    atom rsrc_os,
	 rsrc_vad
    
    fn=open(path,"wb")
    if fn<2 then 
	PE_last_error("write_PE_ex : cant open "&path)
	return 0 
    end if
    
    nsect=length(sections)
    hdr_info=get_info_defs(hdr_info)
    sections=process_sections(sections,hdr_info)
    
    os= PE_size_of(IMAGE_DOS_HEADER)
	+length(STUB_PROG)
	+length(hdr_info[hi_StubMessage])+1
    
    os=PE_round_to(os,hdr_info[hi_FileAlignment])
    
    s=PE_blank_struct(IMAGE_DOS_HEADER)
    s=PE_set_member(s,IMAGE_DOS_HEADER,"e_magic"  ,IMAGE_DOS_SIGNATURE)
    s=PE_set_member(s,IMAGE_DOS_HEADER,"e_cblp"   ,64)
    s=PE_set_member(s,IMAGE_DOS_HEADER,"e_cp"     ,floor(os/64)+not integer(os/64))
    s=PE_set_member(s,IMAGE_DOS_HEADER,"e_cparhdr",floor(os/64)+not integer(os/64))
    s=PE_set_member(s,IMAGE_DOS_HEADER,"e_minalloc",#0F)
    s=PE_set_member(s,IMAGE_DOS_HEADER,"e_maxalloc",#FFFF)
    s=PE_set_member(s,IMAGE_DOS_HEADER,"e_sp"      ,#B8)
    s=PE_set_member(s,IMAGE_DOS_HEADER,"e_lfarlc"  ,#40)
    s=PE_set_member(s,IMAGE_DOS_HEADER,"e_lfanew"  ,os)
    
    PE_write_struct_to_file(fn,IMAGE_DOS_HEADER,s)
    
    puts(fn,STUB_PROG&hdr_info[hi_StubMessage]&#24)
    PE_write_to(fn,os)
    puts(fn,int_to_bytes(IMAGE_NT_SIGNATURE))
    
    s=PE_blank_struct(IMAGE_FILE_HEADER)
    s=PE_set_member(s,IMAGE_FILE_HEADER,"Machine"        ,IMAGE_FILE_MACHINE_I386)
    s=PE_set_member(s,IMAGE_FILE_HEADER,"NumberOfSections",nsect)
    s=PE_set_member(s,IMAGE_FILE_HEADER,"TimeDateStamp"  ,hdr_info[hi_TimeDateStamp])
    s=PE_set_member(s,IMAGE_FILE_HEADER,"SizeOfOptionalHeader",#E0)
    s=PE_set_member(s,IMAGE_FILE_HEADER,"Characteristics"     ,hdr_info[hi_Characteristics])
    PE_write_struct_to_file(fn,IMAGE_FILE_HEADER,s)

    os=PE_round_to(where(fn)+PE_size_of(IMAGE_OPTIONAL_HEADER)
		    +8*IMAGE_NUMBEROF_DIRECTIORY_ENTRIES
		    +nsect*PE_size_of(IMAGE_SECTION_HEADER),
				    hdr_info[hi_FileAlignment])
		    
    s=PE_blank_struct(IMAGE_OPTIONAL_HEADER)
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"Magic",               #10B)
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"MajorLinkerVersion"   ,0)
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"MajorLinkerVersion"   ,1)
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"SizeOfCode",get_code_size(sections,hdr_info))
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"SizeOfInitializedData",get_init_size(sections,hdr_info))
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"SizeOfUninitializedData",get_uninit_size(sections,hdr_info))
    if length(sections[1]) then
	s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"AddressOfEntryPoint",sections[1][1][sh_RVA]+hdr_info[hi_EntryPointOffset])
    end if
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"BaseOfCode"   ,hdr_info[3])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"BaseOfData"   ,get_data_base(sections))
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"ImageBase"    ,hdr_info[hi_ImageBase])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"SectionAlignment" ,hdr_info[hi_SectionAlignment])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"FileAlignment"    ,hdr_info[hi_FileAlignment])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"MajorOperatingSystemVersion",hdr_info[hi_MajorOperatingSystemVersion])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"MinorOperatingSystemVersion",hdr_info[hi_MinorOperatingSystemVersion])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"MajorImageVersion"    ,hdr_info[hi_MajorImageVersion])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"MinorImageVersion"    ,hdr_info[hi_MinorImageVersion])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"MajorSubsystemVersion",hdr_info[hi_MajorSubsystemVersion])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"MinorSubsystemVersion",hdr_info[hi_MinorSubsystemVersion])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"SizeOfImage"  ,get_image_size(sections,hdr_info))
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"SizeOfHeaders",os)
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"Subsystem"    ,hdr_info[hi_Subsystem])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"SizeOfStackReserve",hdr_info[hi_SizeOfStackReserve])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"SizeOfStackCommit" ,hdr_info[hi_SizeOfStackCommit])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"SizeOfHeapReserve" ,hdr_info[hi_SizeOfHeapReserve])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"SizeOfHeapCommit"  ,hdr_info[hi_SizeOfHeapCommit])
    s=PE_set_member(s,IMAGE_OPTIONAL_HEADER,"NumberOfRvaAndSizes",IMAGE_NUMBEROF_DIRECTIORY_ENTRIES)
    PE_write_struct_to_file(fn,IMAGE_OPTIONAL_HEADER,s)
    
    t=get_dir_links(sections)
    for x=1 to IMAGE_NUMBEROF_DIRECTIORY_ENTRIES do
	s=PE_blank_struct(IMAGE_DATA_DIRECTORY)
	if sequence(t[x]) then
	    s=PE_set_member(s,IMAGE_DATA_DIRECTORY,"VirtualAddress",t[x][6])
	    s=PE_set_member(s,IMAGE_DATA_DIRECTORY,"Size",get_section_size(t[x]))
	end if
	PE_write_struct_to_file(fn,IMAGE_DATA_DIRECTORY,s)
    end for

    if sequence(t[3]) then
	rsrc_os=t[3][sh_FileOffset]+os
	rsrc_vad=t[3][sh_RVA]
    else
	rsrc_os=-1
	rsrc_vad=-1
    end if

    for y=1 to length(sections) do
	for x=1 to length(sections[y]) do
	    s=PE_blank_struct(IMAGE_SECTION_HEADER)
	    s=PE_set_member(s,IMAGE_SECTION_HEADER,"Name"          ,
					    sections[y][x][sh_Name])
	    s=PE_set_member(s,IMAGE_SECTION_HEADER,"VirtualSize"   ,
			PE_round_to(get_section_size(sections[y][x]),hdr_info[hi_SectionAlignment]))
	    s=PE_set_member(s,IMAGE_SECTION_HEADER,"VirtualAddress",
					    sections[y][x][sh_RVA])
	    s=PE_set_member(s,IMAGE_SECTION_HEADER,"SizeOfRawData" ,
			PE_round_to(get_section_vsize(sections[y][x]),hdr_info[hi_FileAlignment]))
	    if get_section_size(sections[y][x])  then
		s=PE_set_member(s,IMAGE_SECTION_HEADER,"PointerToRawData",
					    sections[y][x][sh_FileOffset]+os)
	    end if
	    s=PE_set_member(s,IMAGE_SECTION_HEADER,"Characteristics",
					    sections[y][x][sh_Characteristics])
	    
	    PE_write_struct_to_file(fn,IMAGE_SECTION_HEADER,s)
	end for
    end for

    for y=1 to length(sections) do
	for x=1 to length(sections[y]) do
	    PE_write_to(fn,os+sections[y][x][sh_FileOffset])
	    if write_section(fn,sections[y][x]) then
		close(fn)
		PE_last_error("write_PE_ex : write_section failed "&path)
		return 0
	    end if
	end for
    end for
    PE_write_to(fn,PE_round_to(where(fn),hdr_info[hi_FileAlignment]))

    close(fn)
    return {rsrc_os,rsrc_vad}
end function

function find_entry_os(sequence pe)
    atom os,t
    os=PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"AddressOfEntryPoint")
    for x=1 to length(pe[5]) do
	t=PE_get_member(pe[5][x],IMAGE_SECTION_HEADER,"VirtualAddress") 
	if os<t+PE_get_member(pe[5][x],IMAGE_SECTION_HEADER,"SizeOfRawData") and os>=t then
	    return {x,os-t}
	end if
    end for
    PE_last_error("find_entry_os : not found ")
    return {-1,-1}
end function

function find_section_dir_entry(sequence pe,integer section)
    atom os
    os=PE_get_member(pe[5][section],IMAGE_SECTION_HEADER,"VirtualAddress")
    for x=1 to length(pe[4]) do
	if PE_get_member(pe[4][x],IMAGE_DATA_DIRECTORY,"VirtualAddress")=os then
	    return x
	end if
    end for
    return 0
end function

function find_section_data(sequence pe,integer section)
    atom os
    os=PE_get_member(pe[5][section],IMAGE_SECTION_HEADER,"PointerToRawData")
    for x=1 to length(pe[4]) do
	if PE_get_member(pe[4][x],IMAGE_DATA_DIRECTORY,"VirtualAddress")=os then
	    return x
	end if
    end for
    return 0
end function

function order_elements(sequence s,integer n)
    sequence r
    r={}
    r=append(r,s[n])
    for x=1 to length(s) do
	if x!=n then r=append(r,s[x]) end if
    end for
    return r
end function

function PE_load_fmt(sequence path)
    sequence hdr_info,sec
    object eos,pe
    pe=PE_load_header(path)
    if atom(pe) then
	PE_last_error("PE_load_fmt : PE_load_header failed")
	return 1
    end if

    hdr_info=repeat(0,18)
    eos=find_entry_os(pe)
    if eos[1]!=1 then
	pe[5]=order_elements(pe[5],eos[1])
    end if
    eos=eos[2]
    hdr_info[hi_TimeDateStamp]      =PE_get_member(pe[2],IMAGE_FILE_HEADER,"TimeDateStamp")
    hdr_info[hi_Characteristics]    =PE_get_member(pe[2],IMAGE_FILE_HEADER,"Characteristics")
    hdr_info[hi_SectionAlignment]   =PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"SectionAlignment")
    hdr_info[hi_FileAlignment]      =PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"FileAlignment")
    hdr_info[hi_MajorOperatingSystemVersion]=PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"MajorOperatingSystemVersion")
    hdr_info[hi_MinorOperatingSystemVersion]=PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"MinorOperatingSystemVersion")
    hdr_info[hi_MajorImageVersion]  =PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"MajorImageVersion")
    hdr_info[hi_MinorImageVersion]  =PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"MinorImageVersion")
    hdr_info[hi_MajorSubsystemVersion]=PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"MajorSubsystemVersion")
    hdr_info[hi_MinorSubsystemVersion]=PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"MinorSubsystemVersion")
    hdr_info[hi_Subsystem]          =PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"Subsystem")
    hdr_info[hi_SizeOfStackReserve] =PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"SizeOfStackReserve")
    hdr_info[hi_SizeOfStackCommit]  =PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"SizeOfStackCommit")
    hdr_info[hi_SizeOfHeapReserve]  =PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"SizeOfHeapReserve")
    hdr_info[hi_SizeOfHeapCommit]   =PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"SizeOfHeapCommit")
    hdr_info[hi_EntryPointOffset]   =eos
    hdr_info[hi_ImageBase]          =PE_get_member(pe[3],IMAGE_OPTIONAL_HEADER,"ImageBase")
    hdr_info[hi_StubMessage]        ="this is a Windows NT windowed executable"&#D&#A
    sec=repeat(repeat(0,4),length(pe[5]))
    for x=1 to length(sec) do
	sec[x][sh_Name]=PE_get_member(pe[5][x],IMAGE_SECTION_HEADER,"Name")
	sec[x][sh_Characteristics]=PE_get_member(pe[5][x],IMAGE_SECTION_HEADER,"Characteristics")
	sec[x][sh_DirEntry]=find_section_dir_entry(pe,x)
	sec[x][sh_Data]=prepend(find_section_loc(pe,x),path)
	if not sec[x][sh_Data][2] then
	    sec[x][sh_Data]={"",0,PE_get_member(pe[5][x],IMAGE_SECTION_HEADER,"SizeOfRawData")}
	end if
    end for
    return {hdr_info,sec}
end function

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

global function rsrc_get_raw(sequence path,sequence pe,sequence entry)
    integer nsec,fn
    sequence ob,r
    nsec=PE_lookup_section(pe,".rsrc")
    if nsec then
	ob=entry[RSRC_Object]+
	    {PE_get_member(pe[5][nsec],IMAGE_SECTION_HEADER,"PointerToRawData"),0}
	fn=open(path,"rb")
	r={}
	if fn>1 then
	    if seek(fn,ob[1]) then end if
	    for x=1 to ob[2] do
		r&=getc(fn)
	    end for
	    close(fn)
	    return r
	end if
    end if
    PE_last_error("rsrc_get_raw : couldnt open "&path)
    return 1
end function

function find_rsrc_type(integer n,sequence rsrc_tree)
    -- reverses back down the tree to find
    -- type from 1st level directory
    if not n then
	return 0
    end if
    if rsrc_tree[n][RSRC_Parent]-1 then
	return find_rsrc_type(rsrc_tree[n][RSRC_Parent],rsrc_tree)
    else
	if atom(rsrc_tree[n][RSRC_Name]) then
	    return rsrc_tree[n][RSRC_Name]
	else
	    -- Case for named type directory
	    return 0
	end if
    end if
end function

function PE_strip_zeros(sequence s)
    sequence r
    r={} for x=1 to length(s) do
    if s[x] then r&=s[x] end if
    end for return r
end function

function PE_make_number(sequence s)
    -- turn a sequence of bytes into an integer
    if length(s)>4 then
	return -1
    end if
    if length(s)<4 then
	s&=repeat(0,4-length(s))
    end if
    return bytes_to_int(s)
end function

function speek(sequence sect,integer ofs,integer len)
    if ofs+len<=length(sect) then
	return sect[ofs+1..ofs+len]
    end if
    PE_last_error("speek : invalid os/len")
    return -1
end function

function read_rsrc_entry(   integer ofs,
			     object name,
			   sequence sect,
			     object parent,
			   sequence rsrc_tree,
			       atom rva           )
    sequence r
    r=repeat(0,number_of_rsrc_fields)
    
    r[RSRC_Name]    =   name
    r[RSRC_Object]  ={  PE_make_number(speek(sect,ofs   ,4))-rva,
			PE_make_number(speek(sect,ofs+4 ,4))   }
    r[RSRC_CodePage]=   PE_make_number(speek(sect,ofs+8 ,4))
    r[RSRC_Reserved]=   PE_make_number(speek(sect,ofs+12,4))
    r[RSRC_Type]    =   find_rsrc_type(parent,rsrc_tree)
    r[RSRC_Parent]  =   parent
    
    return append(rsrc_tree,r)
end function

function read_rsrc_dir(  integer    ofs,
			  object    name,
			sequence    raw,
			  object    parent,
			sequence    rsrc_tree,
			    atom    rva         )
    sequence r
    atom eofs

    r=repeat(0,number_of_rsrc_fields)
    r[RSRC_Name]            =name
    r[RSRC_Characteristics] =PE_make_number(speek(raw,ofs,4)) ofs+=4
    r[RSRC_TimeDateStamp]   =PE_make_number(speek(raw,ofs,4)) ofs+=4
    r[RSRC_MajorVersion]    =PE_make_number(speek(raw,ofs,2)) ofs+=2
    r[RSRC_MinorVersion]    =PE_make_number(speek(raw,ofs,2)) ofs+=2
    r[RSRC_NumberOfEntries] =PE_make_number(speek(raw,ofs,2)) ofs+=2
    r[RSRC_NumberOfEntries]+=PE_make_number(speek(raw,ofs,2)) ofs+=2
    r[RSRC_Parent]          =parent
    r[RSRC_Type]            =RT_DIRECTORY

    if find(-1,r) then return 1 end if
    rsrc_tree=append(rsrc_tree,r)
    parent=length(rsrc_tree)

    for x=1 to r[RSRC_NumberOfEntries] do
	name    =bytes_to_int(speek(raw,ofs,4)) ofs+=4
	eofs    =bytes_to_int(speek(raw,ofs,4)) ofs+=4
	if and_bits(IMAGE_RESOURCE_NAME_IS_STRING,name) then
	    -- named section
	    name-=IMAGE_RESOURCE_NAME_IS_STRING
	    name =PE_strip_zeros(  speek(   
				       raw,
				       name+2,
				       bytes_to_int(speek(raw,name,2)&0&0) 
							*(RSRC_UNICODE+1) ))
	else
	    -- id section
	end if
	if and_bits(IMAGE_RESOURCE_DATA_IS_DIRECTORY,eofs) then
	    -- sub dir
	    rsrc_tree=read_rsrc_dir(    eofs-IMAGE_RESOURCE_DATA_IS_DIRECTORY,
					name,
					raw,
					parent,
					rsrc_tree,
					rva           )
	else
	    -- data entry
	    rsrc_tree=read_rsrc_entry(  eofs,
					name,
					raw,
					parent,
					rsrc_tree,
					rva           )
	end if
    end for
    return rsrc_tree
end function

global function read_rsrc_tree(sequence path,sequence pe,sequence section)
    integer nsec
    atom rva
    nsec=PE_lookup_section(pe,section)
    if nsec then
	section=PE_get_section(path,pe,nsec)
	rva=PE_get_member(pe[5][nsec],IMAGE_SECTION_HEADER,"VirtualAddress")
	return read_rsrc_dir(0,".rsrc",section,0,{},rva)
    end if
    PE_last_error("read_rsrc_tree : section not found")
    return 1
end function

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

function num_entries(integer dirndx,sequence rsrc_tree)
    integer num,nam
    num=0 nam=0
    for x=1 to length(rsrc_tree) do
	if rsrc_tree[x][RSRC_Parent]=dirndx then
	    if integer(rsrc_tree[x][RSRC_Name]) then
		num+=1 else nam+=1
	    end if
	end if
    end for
    return {nam,num}
end function

global function PE_file_length(sequence path)
    integer fn
    atom p
    fn=open(path,"rb")
    if fn<2 then
	PE_last_error("PE_file_length : cant open "&path)
	return -1
    end if
    if seek(fn,-1) then end if
    p=where(fn)
    close(fn)
    return p
end function

function get_rsrc_size(sequence rsrc)
    if sequence(rsrc[RSRC_Object]) then
	if length(rsrc[RSRC_Object])=3 then
	    if rsrc[RSRC_Object][3] then
		return rsrc[RSRC_Object][3]
	    else
		return PE_file_length(rsrc[RSRC_Object][1])
	    end if
	else
	    return length(rsrc[RSRC_Object])
	end if
    end if
    PE_last_error("get_rsrc_size : invalid section def")
    return 0
end function

function fix_parent(    sequence dat,
			sequence rsrc_tree,
			sequence fix,
			integer  n           )
    
    --Write entry in parent of item n
    integer p,os
    sequence num
    
    p=rsrc_tree[n][RSRC_Parent]
    num=num_entries(p,rsrc_tree)
    
    if p then
	if atom(rsrc_tree[n][RSRC_Name]) then
	    -- ID Entry
	    os=fix[p]+PE_size_of(IMAGE_RESOURCE_DIRECTORY)+
			num[1]*8 +rsrc_tree[p][RSRC_EuObject][2]*8+1
	    dat[os..os+3]=int_to_bytes(rsrc_tree[n][RSRC_Name])
	    if rsrc_tree[n][RSRC_Type]=RT_DIRECTORY then
		dat[os+4..os+7]=int_to_bytes(or_bits(fix[n],IMAGE_RESOURCE_DATA_IS_DIRECTORY))
	    else
		dat[os+4..os+7]=int_to_bytes(fix[n])
	    end if
	    rsrc_tree[p][RSRC_EuObject][2]+=1
	else
	    --Named Entry
	    os=fix[p]+PE_size_of(IMAGE_RESOURCE_DIRECTORY)+
			rsrc_tree[p][RSRC_EuObject][1]*8+1

	    dat[os..os+3]=int_to_bytes(or_bits(rsrc_tree[n][RSRC_EuObject][3],IMAGE_RESOURCE_NAME_IS_STRING))
	    if rsrc_tree[n][RSRC_Type]=RT_DIRECTORY then
		dat[os+4..os+7]=int_to_bytes(or_bits(fix[n],IMAGE_RESOURCE_DATA_IS_DIRECTORY))
	    else
		dat[os+4..os+7]=int_to_bytes(fix[n])
	    end if
	    
	    rsrc_tree[p][RSRC_EuObject][1]+=1
	end if
    end if
    return {dat,rsrc_tree}
end function

function make_unicode(sequence s)
    sequence r
    if RSRC_UNICODE then
	r={}
	for x=1 to length(s) do
	    r&=s[x]&0
	end for
	return r
    else
	return s
    end if
end function

global function make_rsrc_dir( sequence rsrc_tree )
    sequence s,
	     r,
	     fix,
	     rvafix,
	     temp
    
    object t

    fix =repeat(0,length(rsrc_tree))
    r={}
    --build dirs
    for x=1 to length(fix) do
	rsrc_tree[x][RSRC_EuObject]={0,0,0} -- reuse for counts
	if rsrc_tree[x][RSRC_Type]=RT_DIRECTORY then
	    s=PE_blank_struct(IMAGE_RESOURCE_DIRECTORY)
	    s=PE_set_member(s,IMAGE_RESOURCE_DIRECTORY,"Characteristics",rsrc_tree[x][RSRC_Characteristics])
	    s=PE_set_member(s,IMAGE_RESOURCE_DIRECTORY,"TimeDateStamp",rsrc_tree[x][RSRC_TimeDateStamp])
	    s=PE_set_member(s,IMAGE_RESOURCE_DIRECTORY,"MajorVersion",rsrc_tree[x][RSRC_MajorVersion])
	    s=PE_set_member(s,IMAGE_RESOURCE_DIRECTORY,"MinorVersion",rsrc_tree[x][RSRC_MinorVersion])
	    t=num_entries(x,rsrc_tree)
	    s=PE_set_member(s,IMAGE_RESOURCE_DIRECTORY,"NumberOfNamedEntries",t[1])
	    s=PE_set_member(s,IMAGE_RESOURCE_DIRECTORY,"NumberOfIdEntries",t[2])
	    fix[x]=length(r)
	    r&=PE_struct_to_bytes(IMAGE_RESOURCE_DIRECTORY,s)
	    r&=repeat(0,(t[1]+t[2])*8)
	end if
    end for

    rvafix={}
    --build entries
    for x=1 to length(fix) do
	if rsrc_tree[x][RSRC_Type]!=RT_DIRECTORY then
	    s=PE_blank_struct(IMAGE_RESOURCE_DATA_ENTRY)
	    s=PE_set_member(s,IMAGE_RESOURCE_DATA_ENTRY,"CodePage",rsrc_tree[x][RSRC_CodePage])
	    s=PE_set_member(s,IMAGE_RESOURCE_DATA_ENTRY,"Reserved",rsrc_tree[x][RSRC_Reserved])
	    if length(rsrc_tree[x][RSRC_Object])=3 or length(rsrc_tree[x][RSRC_Object])=2 then
		s=PE_set_member(s,IMAGE_RESOURCE_DATA_ENTRY,"Size",rsrc_tree[x][RSRC_Object][2])
	    else
		s=PE_set_member(s,IMAGE_RESOURCE_DATA_ENTRY,"Size",length(rsrc_tree[x][RSRC_Object]))
	    end if
	    rvafix&=length(r)
	    fix[x] =length(r)
	    r&=PE_struct_to_bytes(IMAGE_RESOURCE_DATA_ENTRY,s)
	end if
    end for
    --build name strings
    for x=1 to length(fix) do
	if sequence(rsrc_tree[x][RSRC_Name]) then

	    rsrc_tree[x][RSRC_EuObject][3]=length(r)

	    t=int_to_bytes(length(rsrc_tree[x][RSRC_Name]))
	    r&=t[1..2]&make_unicode(rsrc_tree[x][RSRC_Name])
	    --border 4
	    if not integer(length(r)/4) then
		r&=repeat(0,(floor(length(r)/4)+1)*4-length(r))
	    end if
	end if
    end for
    
    --set entry offsets 
    --build {{os,type},rsrc ob} list and rva fix list
    t=length(r)
    s={} 
    for x=1 to length(fix) do
	temp=fix_parent(r,rsrc_tree,fix,x)
	r=temp[1] rsrc_tree=temp[2]
	if rsrc_tree[x][RSRC_Type]!=RT_DIRECTORY then
	    if length(rsrc_tree[x][RSRC_Object])=2 then
		rsrc_tree[x][RSRC_Object]&=0
	    end if
	    s=append(s,{{t,rsrc_tree[x][RSRC_Type]},rsrc_tree[x][RSRC_Object]})
	    r[fix[x]+1..fix[x]+4]=int_to_bytes(t)
	    if length(rsrc_tree[x][RSRC_Object])=3 then
		t+=rsrc_tree[x][RSRC_Object][2]
	    else
		t+=length(rsrc_tree[x][RSRC_Object])
	    end if
	    --border 4
	    if not integer(t/4) then
		t=(floor(t/4)+1)*4
	    end if
	end if
    end for
    return {t,r,rvafix,s}
end function

global function get_rsrc_type_name(object id)
    if atom(id) then
	-- ordinal type
	for x=1 to length(RSRC_TYPES) do
	    if id=RSRC_TYPES[x][2] then
		-- standard NT resource types
		return RSRC_TYPES[x][1]
	    end if
	end for
	PE_last_error("get_rsrc_type_name : Unknown ordinal type")
	return sprintf("Unknown Resource Type: 0x%.2x",id)
    else
	-- named type
	return '*'&id
    end if
end function

global procedure dump_tree(integer fn,sequence tree)
    for x=1 to length(tree) do
	if atom(tree[x][RSRC_Name]) then
	    printf(fn,"\n(%d)%d\n",{x,tree[x][RSRC_Name]})
	else
	    printf(fn,"\n(%d)%s\n",{x,tree[x][RSRC_Name]})
	end if
	printf(fn,"%s\n",{get_rsrc_type_name(tree[x][RSRC_Type])})
	printf(fn,"Parent: %d\n",tree[x][RSRC_Parent])
	printf(fn,"Entries: %d\n",tree[x][RSRC_NumberOfEntries])
	if sequence(tree[x][RSRC_Object]) then
	    if length(tree[x][RSRC_Object])=2 then
		printf(fn,"Offset:0x%x Size:0x%x\n",tree[x][RSRC_Object])
	    end if
	end if
    end for
end procedure

