-------------------------------------------------------------------------------
-- Support for resource.ew PE.e and Plugins ----------------------G.Burke 01/05
-------------------------------------------------------------------------------

include machine.e
global constant

	    RSRC_Plugin_Type       =1,  -- name or standard type ID
	    RSRC_Plugin_Extension  =2,  -- associated file extension
	    RSRC_Plugin_IsGroup    =3,  -- associated file extension
			    -- Routine IDs --
	    RSRC_Plugin_FileToRsrc =4,  -- File format to RSRC format
	    RSRC_Plugin_RsrcToEu   =5,  -- RSRC format to EU object
	    RSRC_Plugin_EuToRsrc   =6,  -- Eu Object to RSRC format
	    RSRC_Plugin_RsrcToFile =7,  -- RSRC format to File

	    -- Predefined Windows Resource Types --
	    
	    RT_CURSOR           =1,
	    RT_BITMAP           =2,
	    RT_ICON             =3,
	    RT_MENU             =4,
	    RT_DIALOG           =5,
	    RT_STRING           =6,
	    RT_FONTDIR          =7,
	    RT_FONT             =8,
	    RT_ACCELERATOR      =9,
	    RT_RCDATA           =10,
	    RT_MESSAGETABLE     =11,

	    RSRC_RT_DIFFERENCE=11,

	    RT_GROUP_CURSOR     =RT_CURSOR +  RSRC_RT_DIFFERENCE,
	    RT_GROUP_ICON       =RT_ICON   +  RSRC_RT_DIFFERENCE,
	    RT_VERSION          =16,
	    RT_DLGINCLUDE       =17,
	    RT_PLUGPLAY         =19,
	    RT_VXD              =20,
	    RT_DIALOG_DATA      =#F0,
	    RT_TOOLBAR          =#F1,
	    --
	    RT_DIRECTORY        =#1000

-------------------------------------------------------------------------------
-----------------------<Error handler>-----------------------------------------
-------------------------------------------------------------------------------

constant PE_error_depth=10

sequence PE_error_list
PE_error_list=repeat({},PE_error_depth)

global procedure PE_last_error(sequence s)
    PE_error_list=prepend(PE_error_list[1..PE_error_depth-1],s)
end procedure

global function PE_get_last_error()
    return PE_error_list[1]
end function

global procedure PE_dump_errors(integer fn)
    for x=1 to length(PE_error_list) do
	puts(fn,PE_error_list[x]&"\n")
    end for
end procedure

-------------------------------------------------------------------------------
-- Routines for handling simple structures                        
-------------------------------------------------------------------------------

-- These routines are designed to produce readable code without a large 
-- number of individual constant declarations. 

function pad_to(sequence s,integer n)
    if length(s)<n then s&=repeat(32,n-length(s)) end if
    return s
end function

global procedure PE_dump_struct(integer fn,sequence s,sequence dat)
    for x=1 to length(s) do
	if sequence(dat[x]) then
	    puts(fn,pad_to(s[x][1],32)&dat[x])
	else
	    printf(fn,pad_to(s[x][1],32)&":0x%8.8x : %d",repeat(dat[x],2))
	end if
	puts(fn,"\n")
    end for
    puts(fn,"\n")
end procedure

global function PE_get_struct_from_file(integer fn,sequence s)
    sequence r,n
    r=repeat(0,length(s))
    for x=1 to length(s) do
	if s[x][2]<=4 then
	    n={0,0,0,0}
	    for b=1 to s[x][2] do
		n[b]=getc(fn)
	    end for
	    r[x]=bytes_to_int(n)
	else
	    n={}
	    for b=1 to s[x][2] do
		n&=getc(fn)
	    end for
	    r[x]=n
	end if
    end for
    return r
end function

global procedure PE_write_struct_to_file(integer fn,sequence struct,sequence dat)
    sequence n
    for x=1 to length(struct) do
	if struct[x][2]>4 then
	    if length(dat[x])<struct[x][2] then
		dat[x]&=repeat(0,struct[x][2]-length(dat[x]))
	    elsif length(dat[x])>struct[x][2] then
		dat[x]=dat[x][1..struct[x][2]]
	    end if
	    puts(fn,dat[x])
	else
	    n=int_to_bytes(dat[x])
	    puts(fn,n[1..struct[x][2]])
	end if
    end for
end procedure

global function PE_struct_to_bytes(sequence struct,sequence dat)
    sequence n,r
    r={}
    for x=1 to length(struct) do
	if struct[x][2]>4 then
	    if length(dat[x])<struct[x][2] then
		dat[x]&=repeat(0,struct[x][2]-length(dat[x]))
	    elsif length(dat[x])>struct[x][2] then
		dat[x]=dat[x][1..struct[x][2]]
	    end if
	    r&=dat[x]
	else
	    n=int_to_bytes(dat[x])
	    r&=n[1..struct[x][2]]
	end if
    end for
    return r
end function

global function PE_bytes_to_struct(sequence struct,sequence dat)
    sequence n,r
    r={}
    dat&=0
    for x=1 to length(struct) do
	n=dat[1..struct[x][2]]
	dat=dat[struct[x][2]+1..length(dat)]
	if length(n)<4 then
	    n&=repeat(0,4-length(n))
	end if
	if length(n)<5 then
	    r=append(r,bytes_to_int(n))
	else
	    r=append(r,n)
	end if
    end for
    return r
end function

global function PE_size_of(sequence s)
    integer r
    r=0
    for x=1 to length(s) do
	r+=s[x][2]
    end for
    return r
end function

global function PE_blank_struct(sequence s)
    return repeat(0,length(s))
end function

global function PE_get_member(sequence data,sequence struct,sequence name)
    --return value of named member
    for x=1 to length(struct) do
	if equal(name,struct[x][1]) then
	    return data[x]
	end if
    end for
    PE_last_error("PE_get_member : member not found")
    return {-1}
end function

global function PE_set_member(sequence data,sequence struct,sequence name,object val)
    --set value of named member
    for x=1 to length(struct) do
	if equal(name,struct[x][1]) then
	    data[x]=val
	    return data
	end if
    end for
    PE_last_error("PE_set_member : member not found")
    return data
end function

